import sys
import os
import io
import logging

from jupyter_dash import JupyterDash

import dash
from dash import html
from dash import Dash
from dash import dcc 
from dash import dash_table
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

import plotly.express as px
from astropy.table import Table, Column
import pandas as pd

from skimage import io as imio 

fig = None


class mossDisputil:

    # Init input tables are IPAC formatted, 
    # pngdir is the directory of the graphic PNG files

    def __init__ (self, imtblpath, spectblpath, pngdir, **kwargs):
#
# { mossDisputil.init
#
        global fig

        #
        # Check arguments
        #

        debug = 0
        debugfname = ''
    
        if ('debugfile' in kwargs):
            
            debug = 1
            debugfname = kwargs.get ('debugfile')

            if (len(debugfname) > 0):
      
                logging.basicConfig (filename=debugfname, \
                    level=logging.DEBUG)
    
                with open (debugfname, 'w') as fdebug:
                    pass
        
        if debug:
            logging.debug ('')
            logging.debug ('Enter mossDisputil.init: imtblpath= ')
            logging.debug (imtblpath)
            logging.debug ('spectblpath= ')
            logging.debug (spectblpath)
            
        if ((imtblpath is None) and (spectblpath is None)):
            msg = 'Both the image table path and spectra table path are ' + \
                'missing: at least one of them is required to run this service.'
            raise Exception (msg)

        if (pngdir is None):
            msg = 'Required input parameter "pngdir" is missing.'
            raise Exception (msg)


            logging.debug (self.tblpath_im)
            logging.debug ('tblpath_spec= ')
            logging.debug (self.tblpath_spec)

        self.pngdir = pngdir
        self.ntbl = 0
        self.tblpath1 = None
        self.tblpath2 = None
        self.tblname1 = None
        self.tblname2 = None
        self.df1 = None
        self.df2 = None
        self.tbldata1 = None
        self.tbldata2 = None
        self.tblcols1 = None
        self.tblcols2 = None
        self.nrows1 = None
        self.nrows2 = None
        self.colname1 = None
        self.colname2 = None
        self.ncols1 = None
        self.ncols2 = None
 
        self.tbl1_display = None
        self.tbl2_display = None
        
        if ((imtblpath is not None) and \
            (spectblpath is not None)):
            
            self.tblpath1 = imtblpath
            self.tblpath2 = spectblpath
            self.ntbl = 2
        
        elif (imtblpath is not None):
            self.tblpath1 = imtblpath
            self.ntbl = 1 
 
        elif (spectblpath is not None):
            self.tblpath1 = spectblpath
            self.ntbl = 1 
        
        if debug:
            logging.debug ('')
            logging.debug (f'ntbl = {self.ntbl:d}')
            logging.debug (f'tblpath1 = {self.tblpath1:s}')
            
            if (self.ntbl > 1):
                logging.debug (f'tblpath2 = {self.tblpath2:s}')
 
        #
        # Read table1 
        #
        ind = self.tblpath1.rfind ('/')
        if (ind != -1):
            self.tblname1 = self.tblpath1[ind+1:]
        else: 
            self.tblname1 = self.tblpath1

        if debug:
            logging.debug ('')
            logging.debug ('here0-0')
      
        self.df1 = None  
        try:
            self.df1 = self.__tblpath2df__ (self.tblpath1)

        except Exception as e:
            msg = 'Failed to read tblpath1:' + str(e)
            raise Exception (msg)

        if debug:
            logging.debug ('')
            logging.debug ('tblpath1 read')
            logging.debug ('df1:')
            logging.debug (self.df1)
       
        self.tbl1_disply = 'inline_block'
 
        self.tbldata1 = self.df1.to_dict ('records') 
        self.tblcols1=[{
            "name": i, 
            "id": i
        } for i in self.df1.columns]
        
        self.nrows1 = len (self.df1.index)
        self.colname1 = list (self.df1.columns)
        self.ncols1 = len(self.colname1)
 

        #
        # If more than one table, read table2 
        #
        if (self.tblpath2 is not None): 
       
            self.tbl2_disply = 'inline_block'
 
            if debug:
                logging.debug ('')
                logging.debug ('read tblpath2')
                logging.debug ('df1:')
            
            ind = self.tblpath2.rfind ('/')
            if (ind != -1):
                self.tblname2 = self.tblpath2[ind+1:]
            else: 
                self.tblname2 = self.tblpath2

            self.df2 = None  
            try:
                self.df2 = self.__tblpath2df__ (self.tblpath2)

            except Exception as e:
                msg = 'Failed to read tblpath1:' + str(e)
                raise Exception (msg)

            if debug:
                logging.debug ('')
                logging.debug ('tblpath2 read')
                logging.debug ('df2:')
                logging.debug (self.df2)
        
            self.tbldata2 = self.df2.to_dict ('records') 
            self.tblcols2=[{
                "name": i, 
                "id": i
            } for i in self.df2.columns]
        
            self.nrows2 = len (self.df2.index)
            self.colname2 = list (self.df2.columns)
            self.ncols2 = len(self.colname2)

        if debug:
            logging.debug ('')
            logging.debug ('pass both tblpath read')
        
        #
        # Placeholder figure for image display space
        #

        fig = {
            "layout": {
                "xaxis": { "visible": False },
                "yaxis": { "visible": False },
                "annotations": [
                    {
                        "text": "No image selected yet.",
                        "xref": "paper",
                        "yref": "paper",
                        "showarrow": False,
                        "font": { "size": 28 }
                    }
                ]
            }
        }

        if debug:
            logging.debug ('')
            logging.debug ('mossDisputil.init ended')
        
#
# }  end mossDisputil.init
#
    

    def setup_mossapp (self):
#
# { mossDisputil.setup_mossapp
#
        debug = 1

        if debug:
            logging.debug ('')
            logging.debug ('Enter mossDisputil.setup_mossapp:')
                 
        global fig

        app = JupyterDash (__name__)

        app.config.suppress_callback_exceptions = True

        #
        # prepare layout elements
        #
        
        if debug:
            logging.debug ('')
            logging.debug ('here0-0')
                 
        tbl1_div = None
        tbl2_div = None
        
        tbl1_div = html.Div (
            id = 'table1_div',
            style= {'display':self.tbl1_display},
            
            children=[       
                html.H3 (
                    id = 'tablename1',
                    children = self.tblname1,
                    style={
                        "width":"80%", 
                        "display":"inline-block",
                        "textAlign":"center", 
                        "color":"#000088"
                    }
                ),

                dash_table.DataTable (
                    id = 'tabledata1',
                    data = self.tbldata1,
                    columns = self.tblcols1,
                    column_selectable="single",
                    selected_columns=[],
                    fixed_rows={'headers': True},
                    style_header={
                        'border':'1px solid #000088',
                        'textAlign':'center',
                        'backgroundColor': '#c2dfff',
                        'fontWeight':'bold',
                        'color': '#114477',
                    }, 
                    style_data={
                        'height': 'auto',
                        'width': 'auto',
                        'textAlign':'left',
                        'backgroundColor': '#ffffff',
                    },
                    style_data_conditional=[
                        {
                            'if': {'row_index': 'odd'},
                            'backgroundColor': '#dddddd',
                        }
                    ],
                    style_cell={
                        'minWidth':100, 'maxWidth':500, 'width': 100,
                        'padding':'0px 5px 0px 5px'
                    },
                    style_table={
                        "border":"2px solid #008888", 
                        "margin":"5px 2px 2px 5px", 
                        "padding":"0px 0px 0px 0px", 
                        "width":"99%",
                        "overflow":"auto"
                    },
                )
            ]
        )

        tbl2_div = html.Div (
            id = 'table2_div',
            style= {'display':self.tbl2_display},
            
            children=[       
                html.H3 (
                    id = 'tablename2',
                    children = self.tblname2,
                    style={
                        "width":"80%", 
                        "display":"inline-block",
                        "textAlign":"center", 
                        "color":"#000088"
                    }
                ),

                dash_table.DataTable (
                    id = 'tabledata2',
                    data = self.tbldata2,
                    columns = self.tblcols2,
                    column_selectable="single",
                    selected_columns=[],
                    fixed_rows={'headers': True},
                    style_header={
                        'border':'1px solid #000088',
                        'textAlign':'center',
                        'backgroundColor': '#c2dfff',
                        'fontWeight':'bold',
                        'color': '#114477',
                    }, 
                    style_data={
                        'height': 'auto',
                        'width': 'auto',
                        'textAlign':'left',
                        'backgroundColor': '#ffffff',
                    },
                    style_data_conditional=[
                        {
                            'if': {'row_index': 'odd'},
                            'backgroundColor': '#dddddd',
                        }
                    ],
                    style_cell={
                        'minWidth':100, 'maxWidth':500, 'width': 100,
                        'padding':'0px 5px 0px 5px'
                    },
                    style_table={
                        "border":"2px solid #008888", 
                        "margin":"5px 2px 2px 5px", 
                        "padding":"0px 0px 0px 0px", 
                        "width":"99%",
                        "overflow":"auto"
                    },
                )
            ]
        )

                 
        # The Display Layout

        app.layout = html.Div (
    
            children=[
                html.Div (
                    id = 'table_disp',
                    style={
                        "width":"35%", 
                        "display":"inline-block",
                    },
                    children=[
                        tbl1_div,
                        tbl2_div
                    ]
                ),
                #----------------------------------------
                html.Div (
                    id = 'image_disp',
                    style={
                        "width":"63%", 
                        "float":"right",
                        "display":"inline-block",
                        "margin":"auto auto auto 5px", 
                        "padding":"0px 0px 0px 0px", 
                        "vertical-align":"top",
                    },
                    children = [
                        html.H3 (
                            id = 'output_selectedcol',
                            #children = '',
                            style={
                                "width":"80%", 
                                "display":"inline-block",
                                "textAlign":"center", 
                                "color":"#000088"
                            }
                        ),
                        dcc.Graph (id='img_graph', figure=fig, \
                            style={'background-color': 'darkgray'})
                    ]     
                )

            ]
        )
        
        if debug:
            logging.debug ('')
            logging.debug ('here0-3')
                 
        @app.callback (

            [Output ('output_selectedcol', 'children'),
             Output ('img_graph',          'figure')],

            [Input ('tabledata1',        'active_cell'), 
             Input ('tabledata1',        'data'),
             Input ('tabledata2',      'active_cell'), 
             Input ('tabledata2',      'data')]
        )

        def getSelectedCell (active_cell1, data1, active_cell2, \
            data2):
            
            debug = 1               

            if debug:
                logging.debug ('')
                logging.debug ('Enter getSelectedCell ')
                logging.debug ('active_cell1: ')
                logging.debug (active_cell1)
                logging.debug ('data1= ')
                logging.debug (data1)
                logging.debug ('active_cell2: ')
                logging.debug (active_cell2)
                logging.debug ('data2= ')
                logging.debug (data2)
 
            
            ctx = dash.callback_context
            
            if debug:
                logging.debug ('')
                logging.debug ('ctx.triggered[0]: ')
                logging.debug (ctx.triggered[0])
            
            if ctx.triggered[0]['value'] is not None:
                row = ctx.triggered[0]['value']['row']          
                col = ctx.triggered[0]['value']['column']          
                colid = ctx.triggered[0]['value']['column_id']          
                table = ctx.triggered[0]['prop_id'].split('.')[0]
                
            if ((active_cell1 is None) and (active_cell2 is None)):
                raise PreventUpdate
            
            #if (table == 'tabledata1'):
            #    datatbl = self.tbldata1
            #elif (table == 'tabledata2'):
            #    datatbl = self.tbldata2
           
            if (table == 'tabledata1'):
                datatbl = data1
            elif (table == 'tabledata2'):
                datatbl = data2
           
            fitsfile = datatbl[row]['fitsfile']
            pngfile  = datatbl[row]['pngfile']
            pngdir   = datatbl[row]['pngdir']
                
            pngpath = pngdir + '/' + pngfile
                       
            try:
                img = imio.imread(pngpath)

                #width  = img.shape[1]
                #height = img.shape[0]

                width  = 800 
                height = 800
                
                fig = px.imshow(img, width=width, height=height, \
                    binary_string=True)

                fig.update_coloraxes(showscale=False)

                fig.update_layout(
                    margin=dict(l=5, r=5, t=5, b=5),
                    paper_bgcolor="darkgray"
                )

                fig.update_xaxes(showticklabels=False)
                fig.update_xaxes(nticks=0)

                fig.update_yaxes(showticklabels=False)
                fig.update_yaxes(nticks=0)

                return (fitsfile, fig)
                #return (pngfile, fig)

            except:
                #return (fitsfile, dash.no_update)
                return (pngfile, dash.no_update)

            if debug:
                logging.debug ('')
                logging.debug ('here0-5: end getSelected')
                 
        if debug:
            logging.debug ('')
            logging.debug ('end setup_mossapp')
                 
        return(app)
#
# } end mossDisputil.setup_mossapp
#
        


    #
    #  read IPAC format metadata table to pandas data frame;
    #  anad for this tools purpose, rearrange data columns for 
    #  display 
    # 
    def __tblpath2df__ (self, tblpath, **kwargs):
#
# { mossDisputil.setup_mossapp
#
        debug = 1 
 
        if debug:
            logging.debug ('')
            logging.debug ('Enter tblpath2df read')
        #
        # First read table to astropy 
        #
        astropytbl = None 
        try:
            astropytbl = Table.read (tblpath, format="ascii.ipac")

        except Exception as e:
            msg = 'Failed to read table to astropy table:' + str(e)
            raise Exception (msg)

        df = astropytbl.to_pandas()
            
        if debug:
            logging.debug ('')
            logging.debug ('tblpath read')
            logging.debug ('df:')
            logging.debug (df)
        
        #
        # rearrange columns: make pngfile the second column, i.e. col 1
        #
        pngcol = df['pngfile']
           
        if debug:
            logging.debug ('')
            logging.debug ('pngcol:')
            logging.debug (pngcol)

        df = df.drop (columns=['pngfile'])
 
        if debug:
            logging.debug ('')
            logging.debug ('df:')
            logging.debug (df)
 
        df.insert (loc=1, column='pngfile', value=pngcol)
 
        if debug:
            logging.debug ('')
            logging.debug ('df:')
            logging.debug (df)
 
        #
        # add pngdir to df1
        #
        df ['pngdir'] = self.pngdir
         
        if debug:
            logging.debug ('')
            logging.debug ('pngdir added df:')
            logging.debug (df)

        return (df)

#
# } end mossDisputil.__tblpath2df
#



